@@ -1,14 +1,22 @@
+# coding: utf8
+from __future__ import print_function
 import logging
+import sys
 import os
 import re
 import webbrowser
 
-from wsgiref.validate import validator as wsgi_validator
+from io import BytesIO, StringIO
+import wsgiref.validate
 from wsgiref.simple_server import make_server
-from cStringIO import StringIO
-from urlparse import urlparse, urlunparse
-from urllib import unquote
-from Cookie import BaseCookie
+try:
+    from http.cookies import BaseCookie
+    from urllib.parse import unquote, urlparse, urlunparse
+except ImportError:
+    from Cookie import BaseCookie  # NOQA
+    from urlparse import urlparse, urlunparse  # NOQA
+    from urllib import unquote  # NOQA
+
 from functools import wraps
 from itertools import chain
 from shutil import copyfileobj
@@ -23,6 +31,8 @@ from pesto.request import Request
 from pesto.wsgiutils import uri_join, make_query
 from pesto.httputils import parse_header
 
+__version__ = '7.3.5'
+
 #: Registry for xpath multimethods
 xpath_registry = {}
 
@@ -30,6 +40,14 @@ xpath_registry = {}
 REGEXP_NAMESPACE = "http://exslt.org/regular-expressions"
 
 
+if sys.version_info > (3, 0):
+    PY3 = True
+    string_types = str
+else:
+    PY3 = False
+    string_types = (unicode, str)
+
+
 class BadStatusError(AssertionError):
     """
     Raised when a non-success HTTP response is found (eg 404 or 500)
@@ -60,7 +78,7 @@ class XPathMultiMethod(object):
         for xpath, func in self.endpoints:
             if el in xpath(el):
                 return func(*args, **kwargs)
-        raise NotImplementedError("Function %s not implemented for element %r"\
+        raise NotImplementedError("Function %s not implemented for element %r"
                                   % (self.__name__, el,))
 
     def register(self, xpath, func):
@@ -68,19 +86,17 @@ class XPathMultiMethod(object):
             XPath('|'.join('../%s' % item for item in xpath.split('|'))),
             func
         ))
-        if not func.func_doc:
+        func_doc = getattr(func, '__doc__', getattr(func, 'func_doc', None))
+        if not func_doc:
             return
 
         # Add wrapped function to the object's docstring
         # Note that ".. comment block" is required to fool rst/sphinx into
         # correctly parsing the indented paragraphs when there is only one
         # registered endpoint.
-        doc = 'For elements matching ``%s``:'\
-              '\n%s\n\n.. comment block\n\n' % (
+        doc = 'For elements matching ``%s``:n%s\n\n.. comment block\n\n' % (
                     xpath,
-                    '\n'.join('    %s' % line
-                              for line in func.func_doc.split('\n'))
-        )
+                  '\n'.join('    %s' % line for line in func_doc.split('\n')))
         self.__doc__ += doc
         self.__name__ = func.__name__
 
@@ -101,6 +117,7 @@ def when(xpath_expr):
         )
         wrapped.__wrapped__ = func
         wrapped.func_doc = multimethod.__doc__
+        wrapped.__doc__ = multimethod.__doc__
         return wrapped
     return when
 
@@ -136,7 +153,7 @@ class ElementWrapper(object):
 
         return '<%s%s>...</%s>' % (
             self.el.tag,
-            ''.join(' %s="%s"' % (key, escapeattrib(value.encode('utf8')))
+            ''.join(' %s="%s"' % (key, escapeattrib(value))
                     for key, value in self.el.attrib.items()),
             self.el.tag,
         )
@@ -370,7 +387,7 @@ class ElementWrapper(object):
         group)
         """
         for el in self.el.xpath(
-            "./ancestor-or-self::form[1]"\
+            "./ancestor-or-self::form[1]"
             "//input[@type='radio' and @name=$name]",
             name=self.el.attrib.get('name', '')
         ):
@@ -412,9 +429,8 @@ class ElementWrapper(object):
         previously selected option.
         """
         if 'multiple' not in \
-           self.el.xpath('./ancestor-or-self::select[1]')[0].attrib:
-            for el in \
-            self.el.xpath("./ancestor-or-self::select[1]//option"):
+                self.el.xpath('./ancestor-or-self::select[1]')[0].attrib:
+            for el in self.el.xpath("./ancestor-or-self::select[1]//option"):
                 if 'selected' in el.attrib:
                     del el.attrib['selected']
 
@@ -453,7 +469,7 @@ class ElementWrapper(object):
         """
         method = self.el.attrib.get('method', 'GET').upper()
         data = self.submit_data(button)
-        path = uri_join_same_server(
+        path = url_join_same_server(
             self.agent.request.request_uri,
             self.el.attrib.get('action', self.agent.request.request_path)
         )
@@ -483,7 +499,7 @@ class ElementWrapper(object):
         form.
         """
         data = []
-        if isinstance(button, basestring):
+        if isinstance(button, string_types):
             button = self(button)
 
         if button and 'name' in button.attrib:
@@ -507,7 +523,7 @@ class ElementWrapper(object):
                     and isinstance(value, tuple):
                 data.append((name, value))
 
-            elif isinstance(value, basestring):
+            elif isinstance(value, string_types):
                 data.append((name, value))
 
             else:
@@ -565,7 +581,7 @@ class ElementWrapper(object):
         else:
             # List of values, eg ``['1', '23', '8']``
             found = set()
-            values = set(unicode(v) for v in values)
+            values = set(str(v) for v in values)
             for el in self.input_group():
                 if el.attrib.get('value') in values:
                     el.attrib['checked'] = ""
@@ -573,8 +589,8 @@ class ElementWrapper(object):
                 elif 'checked' in el.attrib:
                     del el.attrib['checked']
             if found != values:
-                raise AssertionError("Values %r not present"\
-                                     " in checkbox group %r" %\
+                raise AssertionError("Values %r not present"
+                                     " in checkbox group %r" %
                                      (values - found,
                                       self.el.attrib.get('name')))
 
@@ -587,7 +603,7 @@ class ElementWrapper(object):
         button in the group with the given value and checking it.
         """
         if value is not None:
-            value = unicode(value)
+            value = str(value)
         found = False
         for el in self.el.xpath(
             "./ancestor-or-self::form[1]//input[@type='radio' and @name=$n]",
@@ -599,8 +615,8 @@ class ElementWrapper(object):
             elif 'checked' in el.attrib:
                 del el.attrib['checked']
         if value is not None and not found:
-            raise AssertionError("Value %r not present"\
-                                 " in radio button group %r" %\
+            raise AssertionError("Value %r not present"
+                                 " in radio button group %r" %
                                  (value, self.el.attrib.get('name')))
         return self
 
@@ -610,7 +626,7 @@ class ElementWrapper(object):
         Set the value of a textarea control
         """
         if value is not None:
-            value = unicode(value)
+            value = str(value)
         self.el.text = value
         return self
 
@@ -639,7 +655,7 @@ class ElementWrapper(object):
             except KeyError:
                 pass
         else:
-            self.value = unicode(value)
+            self.value = str(value)
         return self
 
     @when("select[@multiple]")  # NOQA
@@ -657,7 +673,7 @@ class ElementWrapper(object):
             ]
 
         found = set()
-        values = set(unicode(v) for v in values)
+        values = set(str(v) for v in values)
         for opt in options:
             value = opt.attrib.get('value', opt.text)
             if value in values:
@@ -680,7 +696,7 @@ class ElementWrapper(object):
         :param values: list of values to be selected
         """
         if value is not None:
-            value = unicode(value)
+            value = str(value)
         found = False
         for opt in self.el.xpath('.//option'):
             if opt.attrib.get('value', opt.text) == value:
@@ -689,21 +705,25 @@ class ElementWrapper(object):
             elif 'selected' in opt.attrib:
                 del opt.attrib['selected']
         if not found and value is not None:
-            raise AssertionError("Value %r not present in select %r" %\
+            raise AssertionError("Value %r not present in select %r" %
                                  (value, self.el.attrib.get('name')))
         return self
 
     def html(self):
         """
         Return an HTML representation of the element
+
+        :rtype: unicode string
         """
-        return tostring(self.el)
+        return tostring(self.el, encoding='unicode')
 
     def pretty(self):
         """
         Return an pretty-printed string representation of the element
+
+        :rtype: unicode string
         """
-        return tostring(self.el, pretty_print=True)
+        return tostring(self.el, encoding='unicode', pretty_print=True)
 
     def striptags(self):
         r"""
@@ -832,7 +852,9 @@ class TestAgent(object):
         - ``request``, ``response`` - the `Pesto <http://pesto.redgecko.org/>`_
           request and response objects associated with the last WSGI request.
 
-        - ``body`` - the body response as a string
+        - ``body`` - the body response as a bytes object
+
+        - ``body_decoded`` - the body response decoded into a string
 
         - ``lxml`` - the lxml representation of the response body (only
            applicable for HTML responses)
@@ -846,7 +868,11 @@ class TestAgent(object):
     """
 
     response_class = MockResponse
+
+    default_charset = 'UTF-8'
+
     _lxml = None
+    _body = None
 
     environ_defaults = {
         'SCRIPT_NAME': "",
@@ -866,13 +892,15 @@ class TestAgent(object):
 
     def __init__(self, app, request=None, response=None, cookies=None,
                  history=None, validate_wsgi=True, host='127.0.0.1',
-                 port='8080', loglevel=None, logger=None):
+                 port='8080', loglevel=None, logger=None,
+                 environ_defaults=None,
+                 close_response=True):
         """
         Initialize the ``TestAgent`` object.
 
         :param app: The WSGI app under test
         :param validate_wsgi: If True, the application under test will be
-                              wrapped in ``wsgiref.validate.wsgi_validator``
+                              wrapped in ``wsgiref.validate.validator``
                               middleware.
         :param host: The host to use for the WSGI environ ``SERVER_NAME`` key
         :param port: The port to use for the WSGI environ ``SERVER_PORT`` key
@@ -880,53 +908,78 @@ class TestAgent(object):
                          ``None`` means no logging
         :param logger: A ``logging.Logger`` object. If supplied ``loglevel``
                        will be ignored.
+        :param close_response: If ``True`` the response iterator will be
+                               read from and closed immediately. If
+                               ``False``, it is up to the caller to handle
+                               closing the WSGI iterator at the end of the
+                               request.
 
         The following parameters are intended for internal use only:
 
+        :param environ_defaults: A list of default WSGI environ values
         :param history: The history list
-        :param cookies: The cookie store to use for the TestAgent.
+        :param cookie: The cookie store to use for the TestAgent.
         :param request: The ``pesto.request.Request`` object associated with
                         the current request
         :param response: The ``pesto.request.Response`` object associated with
                          the current request
         """
-        self.validate_wsgi = validate_wsgi
-        self.app = app
+
+        if logger is not None:
+            logger = logger
+        elif loglevel is not None:
+            logger = logging.getLogger(self.__class__.__name__)
+            logger.setLevel(loglevel)
+        else:
+            logger = None
+
+        self.options = {
+            'host': host,
+            'port': port,
+            'validate_wsgi': validate_wsgi,
+            'logger': logger,
+            'close_response': close_response,
+        }
+
         self.request = request
         self.response = response
-        self.host = host
-        self.port = port
 
-        if self.validate_wsgi:
-            app = wsgi_validator(app)
+        #: The original wsgi application
+        self.app = app
 
-        # Stores file upload field values in forms
-        self.file_uploads = {}
+        if self.options['validate_wsgi']:
+            self.validated_app = wsgiref.validate.validator(app)
+        else:
+            self.validated_app = app
 
         if cookies:
             self.cookies = cookies
         else:
             self.cookies = BaseCookie()
-        if response:
-            self.cookies.update(parse_cookies(response))
+
         if history:
             self.history = history
         else:
             self.history = []
 
-        self.environ_defaults = self.__class__.environ_defaults.copy()
+        if response:
+            _, opts = parse_header(response.get_header('Content-Type'))
+            self.charset = opts.get('charset', self.default_charset)
+            self.cookies.update(parse_cookies(response))
+            if self.options['close_response']:
+                self._read_response()
+        else:
+            self.charset = self.default_charset
+
+        # Stores file upload field values in forms
+        self.file_uploads = {}
+
+        self.environ_defaults = (environ_defaults or self.environ_defaults)\
+                                    .copy()
         self.environ_defaults.update({'SERVER_NAME': host,
                                       'SERVER_PORT': str(port)})
 
-        if logger is not None:
-            self.logger = logger
-        elif loglevel is not None:
-            self.logger = logging.getLogger(self.__class__.__name__)
-            self.logger.setLevel(loglevel)
-        else:
-            self.logger = None
-
-    def make_environ(self, REQUEST_METHOD='GET', PATH_INFO='', wsgi_input='',
+    def make_environ(self, REQUEST_METHOD='GET', PATH_INFO='', wsgi_input=b'',
                      **kwargs):
         """
         Return a dictionary suitable for use as the WSGI environ.
@@ -938,17 +991,17 @@ class TestAgent(object):
         SCRIPT_NAME = kwargs.pop('SCRIPT_NAME',
                                  self.environ_defaults["SCRIPT_NAME"])
 
-        if SCRIPT_NAME and SCRIPT_NAME[-1] == "/":
+        if SCRIPT_NAME and SCRIPT_NAME[-1] == '/':
             SCRIPT_NAME = SCRIPT_NAME[:-1]
-            PATH_INFO = "/" + PATH_INFO
+            PATH_INFO = '/' + PATH_INFO
 
         if '?' in PATH_INFO:
             if 'QUERY_STRING' in kwargs:
-                raise AssertionError("QUERY_STRING specified both in "\
-                                     "PATH_INFO and as argument to "\
+                raise AssertionError("QUERY_STRING specified both in "
+                                     "PATH_INFO and as argument to "
                                      "make_environ")
             PATH_INFO, querystring = PATH_INFO.split('?', 1)
-            kwargs['QUERY_STRING'] = querystring
+            kwargs['QUERY_STRING'] = unicode_to_environ(querystring)
 
         assert re.match(r'^[/A-Za-z0-9\-._~!$/&\'()*+,;=:@%]*$', PATH_INFO), \
             "Path info not URL encoded"
@@ -956,15 +1009,34 @@ class TestAgent(object):
         assert not re.search(r'%(?![A-F0-9]{2})', PATH_INFO), \
             "Path info not URL encoded"
 
+        # Unquote requires a string argument
+        if isinstance(PATH_INFO, bytes):
+            PATH_INFO = PATH_INFO.decode('ascii')
+        if isinstance(SCRIPT_NAME, bytes):
+            SCRIPT_NAME = SCRIPT_NAME.decode('ascii')
+
+        # 'caf%C3%A9' -> 'caf\xc3\xa9'
         PATH_INFO = unquote(PATH_INFO)
+        SCRIPT_NAME = unquote(SCRIPT_NAME)
+
+        # Python 3 unquotes to unicode correctly but Python 2 gets it wrong
+        if not PY3:
+            # 'caf\xc3\xa0' -> b'caf\xc3\xa9' -> 'café'
+            PATH_INFO = PATH_INFO.encode('latin1').decode(self.charset)
+            SCRIPT_NAME = SCRIPT_NAME.encode('latin1').decode(self.charset)
+
+        PATH_INFO = unicode_to_environ(PATH_INFO)
+        SCRIPT_NAME = unicode_to_environ(SCRIPT_NAME)
 
         environ = self.environ_defaults.copy()
         environ.update(kwargs)
         for key, value in kwargs.items():
+            if '.' not in key:
+                value = unicode_to_environ(value)
             environ[key.replace('wsgi_', 'wsgi.')] = value
 
-        if isinstance(wsgi_input, basestring):
-            wsgi_input = StringIO(wsgi_input)
+        if isinstance(wsgi_input, bytes):
+            wsgi_input = BytesIO(wsgi_input)
 
         environ.update({
             'REQUEST_METHOD': REQUEST_METHOD,
@@ -990,19 +1062,15 @@ class TestAgent(object):
         Return a TestAgent object representing the new state resulting from the
         request.
 
-        environ
-            WSGI environ to be used for the request
+        :param environ: WSGI environ to be used for the request
 
-        follow
-            If false, redirect responses will not be followed
+        :param follow: If false, redirect responses will not be followed
 
-        history
-            If true, a new entry will be added to the history attribute for the
-            resulting TestAgent object
+        :param history: If true, a new entry will be added to the history
+                        attribute for the resulting TestAgent object
 
-        check_status
-            If true, any non success HTTP status code will result in an
-            AssertionError being raised
+        :param check_status: If true, any non success HTTP status code will
+                             result in an AssertionError being raised
         """
         path = environ['SCRIPT_NAME'] + environ['PATH_INFO']
         environ['HTTP_COOKIE'] = '; '.join(
@@ -1016,22 +1084,32 @@ class TestAgent(object):
         else:
             history = self.history
 
-        if self.logger:
-            self.logger.info("%s %s", environ['REQUEST_METHOD'],
-                                      Request(environ).request_uri)
+        if self.options['logger']:
+            self.options['logger'].info("%s %s", environ['REQUEST_METHOD'],
+                                        Request(environ).url)
             if environ['HTTP_COOKIE']:
-                self.logger.debug("Cookie: %s", environ['HTTP_COOKIE'])
+                self.options['logger'].debug("Cookie: %s",
+                                             environ['HTTP_COOKIE'])
             postdata = environ['wsgi.input'].getvalue()
             environ['wsgi.input'].seek(0)
             if postdata:
-                self.logger.info("wsgi.input: %r", postdata)
+                self.options['logger'].info("wsgi.input: %r", postdata)
 
-        response = self.response_class.from_wsgi(self.app, environ,
+        response = self.response_class.from_wsgi(self.validated_app, environ,
                                                  self.start_response)
-        if self.logger:
-            self.logger.info('Response: %s', response.status)
+        agent = self.__class__(
+            self.app,
+            Request(environ),
+            response,
+            self.cookies,
+            history,
+            environ_defaults=self.environ_defaults,
+            **self.options
+        )
+        if self.options['logger']:
+            self.options['logger'].info('Response: %s', response.status)
             for name, value in response.headers:
-                self.logger.debug('Response: %s: %s', name, value)
+                self.options['logger'].debug('Response: %s: %s', name, value)
 
         if check_status and response.status[0] not in '23':
             raise BadStatusError("%s %r returned HTTP status %r" % (
@@ -1040,56 +1118,39 @@ class TestAgent(object):
                 response.status
             ))
 
-        agent = self.__class__(
-            self.app,
-            Request(environ),
-            response,
-            self.cookies,
-            history,
-            validate_wsgi=self.validate_wsgi,
-            host=self.host,
-            port=self.port,
-            logger=self.logger,
-        )
         if follow:
             return agent.follow_all()
         return agent
 
-    def get(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
-            history=True, check_status=True, **kwargs):
+    def _get(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
+             history=True, check_status=True, method='GET', **kwargs):
         """
         Make a GET request to the application and return the response.
 
-        PATH_INFO
+        :param PATH_INFO:
             The path to request from the application. This must be URL encoded.
             As a convenience, query string data can be appended to this.
             Example value: ``agent.get('/my%20script?param=1')``
 
-        data
-            Query string data to be appended to the request, must be either a
-            dict or list of ``(name, value)`` tuples.
+        :param data: Query string data to be appended to the request, must be
+                     either a dict or list of ``(name, value)`` tuples.
 
-        charset
-            Encoding used for any unicode values encountered in ``data``
+        :charset: Encoding used for any unicode values encountered in ``data``
 
-        follow
-            If false, redirect responses will not be followed
+        :follow: If false, redirect responses will not be followed
 
-        history
-            If true, a new entry will be added to the history attribute for the
-            resulting TestAgent object
-
-        check_status
-            If true, any non success HTTP status code will result in an
-            AssertionError being raised
+        :history: If true, a new entry will be added to the history attribute
+                  for the resulting TestAgent object
 
+        :check_status: If true, any non success HTTP status code will result in
+                       an AssertionError being raised
         """
         if data is not None:
             kwargs.setdefault('QUERY_STRING', make_query(data,
                                                          charset=charset))
 
         if self.request:
-            PATH_INFO = uri_join_same_server(self.request.request_uri,
+            PATH_INFO = url_join_same_server(self.request.request_uri,
                                              PATH_INFO)
 
         return self._request(
@@ -1103,50 +1164,53 @@ class TestAgent(object):
             check_status,
         )
 
-    def post(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
-             history=True, check_status=True, **kwargs):
+    def _post(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
+              history=True, check_status=True,
+              content_type='application/x-www-form-urlencoded', method='POST',
+              **kwargs):
         """
         Make a POST request to the application and return the response.
 
-        PATH_INFO
-            The path to request from the application. This must be URL encoded.
+        :param PATH_INFO: The path to request from the application. This must
+                          be URL encoded.
 
-        data
-            POST data to be sent to the application, must be either a
-            dict or list of ``(name, value)`` tuples.
+        :param data: POST data to be sent to the application. Can be a byte
+                     string of the raw post data, a dict or a list of ``(name,
+                     value)`` tuples.
 
-        charset
-            Encoding used for any unicode values encountered in ``data``
+        :param charset: Encoding used for any unicode values encountered in
+                        ``data``
 
-        follow
-            If false, redirect responses will not be followed
+        :param content_type: The content type header for the posted data. The
+                             default is good for testing form submissions, if
+                             you want to test an API you may need to change
+                             this to something else, eg 'application/json'
 
-        history
-            If true, a new entry will be added to the history attribute for the
-            resulting TestAgent object
-
-        check_status
-            If true, any non success HTTP status code will result in an
-            AssertionError being raised
+        :param follow: If false, redirect responses will not be followed
 
+        :param history: If true, a new entry will be added to the history
+                        attribute for the resulting TestAgent object
 
+        :param check_status: If true, any non success HTTP status code will
+                             result in an AssertionError being raised
         """
         if data is None:
             data = []
 
         if self.request:
-            PATH_INFO = uri_join_same_server(self.request.request_uri,
+            PATH_INFO = url_join_same_server(self.request.request_uri,
                                              PATH_INFO)
 
-        data = make_query(data, charset=charset)
-        wsgi_input = StringIO(data)
+        if not isinstance(data, bytes):
+            data = make_query(data, charset=charset).encode('ASCII')
+        wsgi_input = BytesIO(data)
         wsgi_input.seek(0)
 
         return self._request(
             self.make_environ(
-                'POST',
+                method,
                 PATH_INFO,
-                CONTENT_TYPE="application/x-www-form-urlencoded",
+                CONTENT_TYPE=content_type,
                 CONTENT_LENGTH=str(len(data)),
                 wsgi_input=wsgi_input,
                 **kwargs
@@ -1156,6 +1220,21 @@ class TestAgent(object):
             check_status,
         )
 
+    get = _get
+    post = _post
+
+    def head(self, *args, **kwargs):
+        return self._get(method='HEAD', *args, **kwargs)
+
+    def delete(self, *args, **kwargs):
+        return self._get(method='DELETE', *args, **kwargs)
+
+    def put(self, *args, **kwargs):
+        return self._post(method='PUT', *args, **kwargs)
+
+    def patch(self, *args, **kwargs):
+        return self._post(method='PATCH', *args, **kwargs)
+
     def post_multipart(self, PATH_INFO='/', data=None, files=None,
                        charset='UTF-8', follow=True, history=True,
                        check_status=True, **kwargs):
@@ -1163,7 +1242,8 @@ class TestAgent(object):
         POST a request to the given URI using multipart/form-data encoding.
 
         PATH_INFO
-            The path to request from the application. This must be URL encoded.
+            The path to request from the application. This must be a URL
+            encoded string.
 
         data
             POST data to be sent to the application, must be either a
@@ -1195,10 +1275,10 @@ class TestAgent(object):
             files = []
 
         if self.request:
-            PATH_INFO = uri_join_same_server(self.request.request_uri,
+            PATH_INFO = url_join_same_server(self.request.request_uri,
                                              PATH_INFO)
 
-        boundary = '----------------------------------------BoUnDaRyVaLuE'
+        boundary = b'----------------------------------------BoUnDaRyVaLuE'
 
         def add_headers(key, value):
             """
@@ -1209,17 +1289,16 @@ class TestAgent(object):
                 filename, content_type, data = value
                 headers = [
                     ('Content-Disposition',
-                     'form-data; name="%s"; filename="%s"' % (key, filename)),
+                     'form-data; name="%s"; filename="%s"' %
+                            (key, filename)),
                     ('Content-Type', content_type)
                 ]
                 return headers, data
             else:
-                if isinstance(value, unicode):
+                if not isinstance(value, bytes):
                     value = value.encode(charset)
-                headers = [
-                    ('Content-Disposition',
-                    'form-data; name="%s"' % (key,))
-                ]
+                headers = [('Content-Disposition',
+                            'form-data; name="%s"' % (key,))]
                 return headers, value
 
         items = chain(
@@ -1228,24 +1307,28 @@ class TestAgent(object):
              for k, fname, ctype, data in files),
         )
 
-        CRLF = '\r\n'
-        post_data = StringIO()
-        post_data.write('--' + boundary)
+        CRLF = b'\r\n'
+        post_data = BytesIO()
+        post_data.write(b'--' + boundary)
         for headers, data in items:
             post_data.write(CRLF)
             for name, value in headers:
-                post_data.write('%s: %s%s' % (name, value, CRLF))
+                line = '%s: %s' % (name, value)
+                post_data.write(line.encode('ascii'))
+                post_data.write(CRLF)
             post_data.write(CRLF)
             if hasattr(data, 'read'):
                 copyfileobj(data, post_data)
-            elif isinstance(data, str):
+            elif isinstance(data, bytes):
                 post_data.write(data)
+            elif isinstance(data, string_types):
+                post_data.write(data.encode('ascii'))
             else:
                 for chunk in data:
                     post_data.write(chunk)
             post_data.write(CRLF)
-            post_data.write('--' + boundary)
-        post_data.write('--' + CRLF)
+            post_data.write(b'--' + boundary)
+        post_data.write(b'--' + CRLF)
         length = post_data.tell()
         post_data.seek(0)
         kwargs.setdefault('CONTENT_LENGTH', str(length))
@@ -1253,7 +1336,9 @@ class TestAgent(object):
             self.make_environ(
                 'POST',
                 PATH_INFO,
-                CONTENT_TYPE='multipart/form-data; boundary=%s' % boundary,
+                CONTENT_TYPE=unicode_to_environ(
+                                'multipart/form-data; boundary=' +
+                                boundary.decode('ascii')),
                 wsgi_input=post_data,
                 **kwargs
             ),
@@ -1270,11 +1355,17 @@ class TestAgent(object):
         refilled on the reloaded page, provided that the reloaded page has
         exactly the same fields present in the same order.
         """
-        self.logger and self.logger.debug("Reload: %s",
-                                          self.request.request_path)
+        if self.options['logger']:
+            self.options['logger'].debug("Reload: %s",
+                                         self.request.request_path)
         env = self.request.environ.copy()
         del env['pesto.request']
-        env['wsgi.input'].seek(0)
+
+        wsgi_input = env['wsgi.input']
+        if isinstance(wsgi_input, wsgiref.validate.InputWrapper):
+            wsgi_input = wsgi_input.input
+        wsgi_input.seek(0)
+        env['wsgi.input'] = wsgi_input
         agent = self._request(env, follow=follow, check_status=check_status)
 
         for src, dst in zip(
@@ -1325,9 +1416,36 @@ class TestAgent(object):
         else:
             return super(TestAgent, self).__str__()
 
+    def _read_response(self):
+
+        if not self.response:
+            return
+        if self._body is not None:
+            return
+        try:
+            self._body = b''.join(self.response.content)
+        finally:
+            try:
+                self.response.content.close()
+            except AttributeError:
+                pass
+
+    @property
+    def body_bytes(self):
+        """
+        The body of the response
+
+        :rtype: bytes
+        """
+        if self._body is None:
+            self._read_response()
+        return self._body
+
     @property
     def body(self):
-        return self.response.body
+        if self.body_bytes is None:
+            return None
+        return self.body_bytes.decode(self.charset)
 
     @property
     def lxml(self):
@@ -1344,7 +1462,7 @@ class TestAgent(object):
         """
         Reset the lxml document, abandoning any changes made
         """
-        self._lxml = fromstring(self.response.body)
+        self._lxml = fromstring(self.body)
 
     def find(self, path, namespaces=None, **kwargs):
         """
@@ -1515,10 +1633,7 @@ class TestAgent(object):
             self.response,
             None,
             self.history,
-            validate_wsgi=self.validate_wsgi,
-            host=self.host,
-            port=self.port,
-            logger=self.logger
+            **self.options
         )
         agent.cookies = BaseCookie()
         return agent
@@ -1554,11 +1669,11 @@ class TestAgent(object):
         port = int(self.environ_defaults['SERVER_PORT'])
 
         server = make_server(host, port, PassStateWSGIApp(self))
-        url = self.request.make_uri(netloc='{0}:{1}'.format(host, port))
+        url = self.request.make_url(netloc='{0}:{1}'.format(host, port))
         if open_in_browser:
             webbrowser.open_new_tab(url)
-        print "\nStarting HTTP server on {0}:{1}. "\
-              "Press ctrl-c to exit.".format(host, port)
+        print("\nStarting HTTP server on {0}:{1}. "
+              "Press ctrl-c to exit.".format(host, port))
         try:
             server.serve_forever()
         except KeyboardInterrupt:
@@ -1593,7 +1708,7 @@ class PassStateWSGIApp(object):
 
     def __call__(self, environ, start_response):
         if self.first_request_served:
-            return self.testagent.app(environ, start_response)
+            return self.testagent.validated_app(environ, start_response)
         self.first_request_served = True
         return self.first_request(environ, start_response)
 
@@ -1601,45 +1716,46 @@ class PassStateWSGIApp(object):
         response = self.testagent.response
         mimetype, charset = parse_header(response.get_header('Content-Type'))
         charset = dict(charset).get('charset', 'UTF-8')
+        self.testagent._read_response()
         if is_html(response) and self.testagent._lxml is not None:
             response = response.replace(content=[tostring(self.testagent.lxml,
                                                           encoding=charset)])
         else:
-            response = response.replace(content=[response.body])
+            response = response.replace(content=[self.testagent._body])
 
         for key, morsel in self.testagent.cookies.items():
             response = response.add_header('Set-Cookie', morsel.OutputString())
         return response(environ, start_response)
 
 
-def uri_join_same_server(baseuri, uri):
+def url_join_same_server(baseurl, url):
     """
-    Join two uris which are on the same server. The resulting URI will have the
+    Join two urls which are on the same server. The resulting URI will have the
     protocol and netloc portions removed. If the resulting URI has a different
     protocol/netloc then a ``ValueError`` will be raised.
 
-        >>> from flea import uri_join_same_server
-        >>> uri_join_same_server('http://localhost/foo', 'bar')
+        >>> from flea import url_join_same_server
+        >>> url_join_same_server('http://localhost/foo', 'bar')
         '/bar'
-        >>> uri_join_same_server('http://localhost/foo',
+        >>> url_join_same_server('http://localhost/foo',
         ...                      'http://localhost/bar')
         '/bar'
-        >>> uri_join_same_server('http://localhost/rhubarb/custard/',
+        >>> url_join_same_server('http://localhost/rhubarb/custard/',
         ...                      '../')
         '/rhubarb/'
-        >>> uri_join_same_server('http://localhost/foo',
+        >>> url_join_same_server('http://localhost/foo',
         ...                      'http://example.org/bar')
         Traceback (most recent call last):
           ...
         ValueError: URI links to another server: http://example.org/bar
 
     """
-    uri = uri_join(baseuri, uri)
-    uri = urlparse(uri)
-    if urlparse(baseuri)[:2] != uri[:2]:
-        raise ValueError("URI links to another server: %s" %\
-                         (urlunparse(uri),))
-    return urlunparse((None, None) + uri[2:])
+    url = uri_join(baseurl, url)
+    url = urlparse(url)
+    if urlparse(baseurl)[:2] != url[:2]:
+        raise ValueError("URI links to another server: %s" %
+                         (urlunparse(url),))
+    return urlunparse(('', '') + url[2:])
 
 
 def parse_cookies(response):
@@ -1690,3 +1806,11 @@ def guess_expression_flavor(expr):
     if '@' in expr:
         return 'xpath'
     return 'css'
+
+
+def unicode_to_environ(s, enc='UTF-8'):
+    """
+    Return a unicode string encoded for a WSGI environ value.
+    This implementation good for python2 only
+    """
+    return s.encode(enc, 'surrogateescape')
